import numpy as np


class TestFunctions:
    def __init__(self) -> None:
        """Class contains various test functions as static methods
        """
        pass

    @staticmethod
    def booth(x, y):
        return (x + 2 * y - 7) ** 2 + (2 * x + y - 5) ** 2

    @staticmethod
    def beale(x, y):
        return (
            (1.5 - x + x * y) ** 2
            + (2.25 - x + x * y ** 2) ** 2
            + (2.625 - x + x * y ** 3) ** 2
        )

    @staticmethod
    def matyas(x, y):
        return 0.26 * (x ** 2 + y ** 2) - 0.48 * x * y



import random
import numpy as np


class Crossover:
    def __init__(self) -> None:
        pass

    @staticmethod
    def onepoint(population, fitness, cr=0.25):
        indices = []
        random_numbers = np.random.uniform(0, 1, (len(population)))
        for idx, val in enumerate(random_numbers):
            if val < cr:
                indices.append(idx)
        for idx, val in enumerate(indices):
            cur_index = indices[idx]
            nxt_index = indices[idx + 1] if idx < len(indices) - 1 else indices[0]
            index = random.randint(0, len(population) - 1)
            population[cur_index, index:], population[nxt_index, index:] = (
                population[nxt_index, index:],
                population[cur_index, index:],
            )
        return population

    @staticmethod
    def clone(population, fitness):
        return population



import numpy as np
from tqdm import tqdm


class GeneticAlgorithm:
    def __init__(
        self, population, objective_fn, selection, crossover, mutation
    ) -> None:
        """Class combines various objects like population, selection, crossover and mutation 
        to create a GA object and perform operation
        """
        self.population = population
        self.objective_fn = objective_fn
        self.selection = selection
        self.crossover = crossover
        self.mutation = mutation

    def forward(self, iterations):
        objective, pop = None, None
        pop = self.population
        i = None
        for i in tqdm(range(iterations), ncols=100):
            objective = self.objective_fn(*pop.T)
            if (objective == 0).sum() >= 1:
                break
            fitness = 1 / objective
            pop = self.selection(pop, fitness)
            pop = self.crossover(pop, fitness)
            pop = self.mutation(
                pop,
                low=np.min(self.population),
                high=np.max(self.population),
                dtype=self.population[-1][-1].dtype,
            )
        return i, objective, pop



import numpy as np


class Mutation:
    def __init__(self) -> None:
        pass

    def randompoint(pop, low, high, dtype="int", mr=0.2):
        total_mutations = int(mr * len(pop))
        if dtype == "float":
            dtype = "float"
        elif dtype == "int64":
            dtype = "int"
        values = {
            "int": np.random.randint(low, high, (total_mutations)),
            "float": np.random.uniform(low, high, (total_mutations)),
        }
        mutation_points = np.random.randint(0, len(pop), size=(total_mutations))
        mutation_values = values[dtype]
        _pop = pop.reshape(-1, pop.shape[0] * pop.shape[1])
        for index, point in enumerate(mutation_points):
            _pop[0][point] = mutation_values[index]
        pop = _pop.reshape(pop.shape[0], pop.shape[1])
        return pop



import random
import numpy as np


class Selection:
    def __init__(self) -> None:
        pass

    @staticmethod
    def tournament(population, fitness, tournament_size=2):
        new_pop = []
        number_of_tournaments = len(population)
        for _ in range(number_of_tournaments):
            x = random.randint(0, len(population) - 1)
            y = random.randint(0, len(population) - 1)
            if fitness[x] >= fitness[y]:
                new_pop.append(population[x])
            else:
                new_pop.append(population[y])
        return np.array(new_pop)

    @staticmethod
    def roulette(population, fitness):
        probs = fitness / np.sum(fitness)
        cumsum = np.cumsum(probs)
        new_index = []
        for _ in range(len(population)):
            for idx, val in enumerate(fitness):
                r = random.random()
                if r < val:
                    new_index.append(idx)

        return population[new_index]



import random
import numpy as np


class Population:
    """
    Class to generate population to the Genetic Algorithm

    Parameters:
        low: lowest value to generate population value
        high: highest value to generate population value
        seed: set random seed value
        dtype: type of generated population, [`int`, `float`]
        dist: if `float`, distribution to be selected from [`normal`, `uniform`]
        unique: if selected `int`, unique makes sure each chromosome created has unique values in it
    Returns: None
    """

    def __init__(self, low, high, dtype="int", dist=None, unique=False) -> None:
        super().__init__()
        self.unique = unique
        self.dtype = dtype
        self.dist = dist
        self.high = high
        self.low = low

    def forward(self, pop_size, variables):
        """
        Function creates a population 

        Parameters: 
            pop_size: size of population
            variables: variables to optimize in problem 
        Returns: population in numpy.ndarray
        """
        params = (self.low, self.high, (pop_size, variables))
        assert self.dtype in ["int", "float"]
        if self.dist is not None:
            assert self.dist in ["normal", "uniform"]
        if self.dtype == "int":
            if self.unique:
                sample = set(range(self.low, self.high + 1))
                unique_nums = [
                    random.sample(sample, variables) for _ in range(pop_size)
                ]
                return np.array(unique_nums)
            else:
                return np.random.randint(*params)
        else:
            if self.dist == "uniform":
                return np.random.uniform(*params)
            else:
                return np.random.normal(*params)

    def __call__(self, pop_size, variables):
        """
        call the forward function automatically
        """
        return self.forward(pop_size, variables)
